import numpy as _np
import numba as _numba
import numexpr as _ne
from ..util import fastlen, atleastnd, intersect2d, alignedarray
import itertools as _it
import warnings as _w

try:
    from .autocorrelate3 import autocorrelate3
except ImportError:
    _w.warn('using numpy ft')

    def autocorrelate3(data):
        tmp = _np.fft.rfftn(data[..., :-2])
        tmp *= tmp.conj()
        data[..., :-2] = _np.fft.irfftn(tmp, s=_np.add(data.shape, (0, 0, -2)))
        data[..., -2:] = 0
        return 0


def corr(input, z, sampling=None, verbose=False):
    """
    calculated 3d correlation of 2d input array sampled at distance z using fft.
    assumes center of input to be center of image
    if input is 3d, the result will be the sum along the first dimension.
    Parameters:
    z: Detector distance in pixels
    sampling: q resoltion - None (default) -> use center pixel q size
                          - 0: smallest step in qx and qy
                          - number -> centerpixel / number
                          - 3 numbers -> seperate centerpixel/number in qx,qy,qz
    """

    def _prepare(input, z, sampling):
        """
        transform centered 2d input sampled at distance z to 3d k-space
        """

        y, x = _np.meshgrid(_np.arange(input.shape[1], dtype=_np.float64), _np.arange(input.shape[0], dtype=_np.float64))
        x -= input.shape[0] / 2.0
        y -= input.shape[1] / 2.0
        d = _np.sqrt(x ** 2 + y ** 2 + z ** 2)
        qx, qy, qz = [(k / d * z) for k in (x, y, z)]

        if sampling is not None:
            sampling = _np.array(sampling)
            if sampling.size == 1:
                if sampling == 0:
                    # (more) correct, but slower. should think about correct oversampling
                    sampling = 1 / min(abs(qy[1, 1] - qy[0, 0]), abs(qx[1, 1] - qx[0, 0]))
                qx, qy, qz = ((k * sampling) for k in (qx, qy, qz))
            elif sampling.size == 3:
                qx, qy, qz = ((k * s) for k, s in zip((qx, qy, qz), sampling))
            else:
                raise ValueError('sampling should be None or 1 or 3 numbers')

        qx, qy, qz = [_np.rint(k - _np.min(k)).astype(int, copy=False) for k in (qx, qy, qz)]
        qlenx, qleny, qlenz = [fastlen(2 * (_np.max(k) + 1)) for k in (qx, qy, qz)]
        ret = alignedarray((qlenz, qleny, qlenx + 2), dtype=_np.float64, alignment=64, zero=True)  # additonal padding in qx for inplace fft
        _np.add.at(ret, (qz, qy, qx), input)
        #     ret[kz1,ky1,kx1]=input #only if no double assignment
        return ret

    def _corr(input, z, sampling):
        """
        wrapper for autocorrelate3, removes redundant slices in first dimension
        """
        tmp = _prepare(input, z, sampling)
        err = autocorrelate3(tmp)
        if err:
            raise RuntimeError(f'cython autocorrelations failed with error code {err}')
        return tmp[: tmp.shape[0] // 2, ...]

    if input.ndim == 2:
        if verbose:
            print('.', end=' ', flush=True)
        return _corr(input, z, sampling)
    elif input.ndim == 3:
        s = _prepare(_np.zeros_like(input[0, ...]), z, sampling).shape
        res = _np.zeros((s[0] // 2, s[1], s[2]))
        for n, inp in enumerate(input):
            if verbose:
                print(n, end=' ', flush=True)
            _np.add(res, _corr(inp, z), out=res)
        return res
    else:
        raise TypeError


def unwrap(img):
    """
    unwraps a single correlation result
    """
    return _np.roll(img[..., :-2], shift=(img.shape[1] // 2, (img.shape[2] - 2) // 2), axis=(1, 2))


class correlator_tiles:
    def __init__(self, qs, maskout, meandata):
        """
        3d fft correlator for tiled data
        qs: q vectors of data points, array of shape (T,P,3) with T: non overlapping tiles, P: pixels per tile
        maskout: points on detector that will not be used, array of shape (T,P)
        meandata: mean of data values used for normalisation, array of shape (T,P)
        resolution is fixed at dq=1
        """
        q = atleastnd(qs, 3)
        for a, b in _it.combinations(range(q.shape[0]), 2):
            if len(intersect2d(q[a, :, :], q[b, :, :])):
                _w.warn('qs are assumed to be unique in first dimension!')
        if not q.shape[2] == 3:
            raise ValueError('q should qz,qy,qx in last axis')
        mask = atleastnd(maskout, 2)
        mean = atleastnd(meandata, 2)
        if not q.shape[:2] == mask.shape == mean.shape:
            print(q.shape, mask.shape, mean.shape)
            raise ValueError('shape missmatch')

        q = _np.rint(q - _np.min(q[~maskout], axis=0)).astype(int)
        q[mask] = _np.max(q, axis=(0, 1)) + 2
        qlen = _np.array([fastlen(2 * (_np.max(k) + 1)) for k in q[~mask].T])
        self.qlenz = _np.max(q[~mask][:, 0]) + 1  # unpadded
        self.q = q
        self.mask = _np.copy(mask)
        self.accum = None
        self._tmp = None
        assemblenorm = correlator_tiles._getnorm(q, mask)
        with _np.errstate(divide='ignore'):
            self.invmean = 1 / (mean * assemblenorm)
        self.invmean[~_np.isfinite(self.invmean)] = 0
        self._N = 0
        self.finished = False
        self._qlen = qlen

    def suspend(self):
        """
        free tmp buffer
        """
        self._tmp = None

    def __exit__(self, *args):
        self.suspend()

    def __enter__(self):
        if self._tmp is None:
            tmp = alignedarray((self._qlen[0], self._qlen[1], self._qlen[2] + 2), zero=True)
            self._tmp = tmp
        if self.accum is None:
            accum = _np.zeros((self.qlenz, self._qlen[1], self._qlen[2] + 2))
            _np.subtract(accum, 0, out=accum)  # force allocation
            self.accum = accum
        return self

    def add(self, data):
        """
        does correlation of data and adds to internal accumulator
        data: array of shape T,P with T: non overlapping tiles, P pixel per Tile
        """
        d = atleastnd(data, 2)
        if not d.shape == self.q.shape[:-1]:
            print(d.shape, self.q.shape, flush=True)
            raise ValueError('shape missmatch')
        if self.finished:
            raise GeneratorExit('already finished')
        if self._tmp is None:
            tmp = alignedarray((self._qlen[0], self._qlen[1], self._qlen[2] + 2), zero=True)
            self._tmp = tmp
        else:
            _zero(self._tmp)
        if self.accum is None:
            accum = _np.zeros((self.qlenz, self._qlen[1], self._qlen[2] + 2))
            _np.subtract(accum, 0, out=accum)  # force allocation
            self.accum = accum
        d = d * self.invmean
        d[self.mask] = 0
        _addat(self._tmp, self.q, d)
        err = autocorrelate3(self._tmp)
        if err:
            raise RuntimeError(f'cython autocorrelations failed with error code {err}')
        _np.add(self.accum, self._tmp[: self.qlenz, ...], out=self.accum)
        self._N += 1

    def result(self, finish=False):
        """
        returns result of accumulated correlations
        finish: free accumulator and buffer.
        """
        if self.accum is None:
            return None

        _zero(self._tmp)
        assemblenorm = correlator_tiles._getnorm(self.q, self.mask)
        _addat(self._tmp, self.q, _np.sqrt(self.N) * _np.array(~self.mask, dtype=_np.float64) / assemblenorm)
        err = autocorrelate3(self._tmp)
        if err:
            raise RuntimeError(f'cython autocorrelations failed with error code {err}')

        res = _ne.evaluate(
            'where((norm<(10*N)), nan,accum/norm)',
            local_dict={'N': self._N, 'norm': self._tmp[: self.qlenz, ...], 'nan': _np.nan, 'accum': self.accum},
        )
        if finish:
            self.finished = True
            self.accum = None
            self._tmp = None
        res = unwrap(res)
        return res

    @property
    def shape(self):
        """
        shape of the result
        """
        return self.accum.shape

    @property
    def N(self):
        """
        number of images added
        """
        return self._N

    @staticmethod
    def _getnorm(q, mask):
        """
        returns amount of pixels with same q
        """
        maxq = _np.max(q.reshape(-1, 3), axis=0)
        hist = _np.histogramdd(q.reshape(-1, 3), bins=maxq + 1, range=[[-0.5, mq + 0.5] for mq in maxq], weights=~mask.ravel())[0]
        ret = hist[q.reshape(-1, 3)[:, 0], q.reshape(-1, 3)[:, 1], q.reshape(-1, 3)[:, 2]].reshape(q.shape[:2])
        ret[mask] = 1
        return ret


@_numba.njit(parallel=True)
def _addat(array, ind, input):
    """
    sets array to input at ind, parallel
    """
    nproc = ind.shape[0]
    nel = ind.shape[1]
    for i in _numba.prange(nproc):
        tinp = input[i, ...]
        tind = ind[i, ...]
        for j in range(nel):
            array[tind[j, 0], tind[j, 1], tind[j, 2]] += tinp[j]
    return


@_numba.njit(parallel=True)
def _zero(array):
    """
    set array to zero
    """
    a = array.ravel()
    for i in _numba.prange(len(a)):
        a[i] = 0
    return True


class correlator:
    def __init__(self, maskout, z):
        """
        3d fft correlator
        maskout: points on detector that will not be used, shape should be same as images to correlate
        resolution is fixed at dq=1
        """
        y, x = _np.meshgrid(_np.arange(maskout.shape[1], dtype=_np.float64), _np.arange(maskout.shape[0], dtype=_np.float64))
        x -= maskout.shape[0] / 2.0
        y -= maskout.shape[1] / 2.0
        d = _np.sqrt(x ** 2 + y ** 2 + z ** 2)
        qs = _np.array([(k / d * z) for k in (z, y, x)])
        qs = _np.rint(qs - _np.min(qs, (-1, -2), keepdims=True)).astype(int, copy=False)
        qlen = [fastlen(k) for k in 2 * (_np.max(qs, (-1, -2)) + 1)]
        maxq = _np.max(qs, (-1, -2))
        hist = _np.histogramdd(qs.reshape(3, -1).T, bins=maxq + 1, range=[[-0.5, mq + 0.5] for mq in maxq], weights=~maskout.ravel())[0]
        count = hist[qs[0].ravel(), qs[1].ravel(), qs[2].ravel()].reshape(maskout.shape)
        count[maskout] = _np.inf
        self._countcorrection = 1 / count
        self._mask = maskout
        self._qlen = qlen
        self._qs = qs
        self._tmp = None

    def _corr(self, input, maxqz):
        for image in input:
            _zero(self._tmp)
            _np.add.at(self._tmp, (*self._qs,), image * self._countcorrection)
            err = autocorrelate3(self._tmp)
            if err:
                raise RuntimeError(f"cython autocorrelations failed with error code {err}")
            yield self._tmp[: min(self._tmp.shape[0] // 2, maxqz), ...]

    @property
    def mask(self):
        """
        mask being used
        """
        return self._mask

    def corr(self, input, maxqz=_np.inf):
        """
        correlate one or multiple images
        return view that will be destroyed on next call, should be copied!
        """
        if self._tmp is None:
            self._tmp = alignedarray((self._qlen[0], self._qlen[1], self._qlen[2] + 2), zero=True)
        if isinstance(input, _np.ndarray):
            return next(self._corr((input,), maxqz))
        else:
            return self._corr(input, maxqz)

    def __enter__(self):
        if self._tmp is None:
            self._tmp = alignedarray((self._qlen[0], self._qlen[1], self._qlen[2] + 2), zero=True)
        return self

    def __exit__(self, *args):
        self.suspend()

    def suspend(self):
        self._tmp = None

    @staticmethod
    def unwrap(img):
        """
        unwraps a single correlation result
        """
        return _np.roll(img[..., :-2], shift=(img.shape[1] // 2, (img.shape[2] - 2) // 2), axis=(1, 2))



import numpy as _np
import cupy as _cp
from ..util import fastlen as _fastlen


def corr(input, axes=(-1, -2), norm=False, returngpu=False, **kwargs):
    """
    simple autocorrelation of input along axes (default: last two) using gpu
    axes: axes to correlate along, defaults to last two
    norm: do normalisation along non correlation axes and normalise for pair count
    returngpu: retrun a cupy array
    """

    axes = sorted([input.ndim + a if a < 0 else a for a in axes])
    fftshape = [_fastlen(2 * input.shape[ax]) for ax in axes]
    dinput = _cp.array(input)
    if norm:
        dinput *= 1 / dinput.mean(axis=[i for i in range(input.ndim) if i not in axes] or None)
    ret = _cp.fft.rfftn(dinput, fftshape)
    ret = _cp.abs(ret) ** 2
    ret = _cp.fft.irfftn(ret, axes=axes)
    ret = _cp.fft.fftshift(ret, axes=axes)[
        tuple((Ellipsis, *(slice(ps // 2 - input.shape[ax], ps // 2 + input.shape[ax]) for ax, ps in zip(axes, fftshape))))
    ]
    if norm:
        n = corr(_cp.ones(tuple(input.shape[ax] for ax in axes)), returngpu=True)
        ret /= n
        ret[(...,) + (n < 0.9).nonzero()] = _np.nan
    if not returngpu:
        ret = _cp.asnumpy(ret)
        _cp.get_default_memory_pool().free_all_blocks()
    return ret



import multiprocessing as _mp
import ctypes as _ctypes
import numpy as _np
import h5py as _h5py
import contextlib as _contextlib


# _mp = _multiprocessing.get_context('spawn')


class datasetreader:
    """
    reads in multiple datasets of an h5 in seperate background processes
    """

    def __init__(self, datasets, file=None, sizecache=20, willread=None):
        """
        reader for multiple datasets of same shape and dtype of an h5 file

        :param datasets: Iterable of Datasets or Strings
        :param file: None (if datasets are h5 Datasets) or String, h5py File or Group, List of Strings (same length as datasets)
        :param sizecache: int, size of the cache
        :param willread: bool array, indices where willread is false are promised to never be read, or None
        """
        ds_names = list()
        filenames = list()
        for i, d in enumerate(datasets):
            if isinstance(d, _h5py.Dataset):
                ds_names.append(d.name)
                filenames.append(d.file.filename)
            elif isinstance(d, str):
                ds_names.append(d)
                if isinstance(file, str):
                    filenames.append(file)
                elif isinstance(file, list) and len(file) == len(datasets):
                    filenames.append(file[i])
                elif isinstance(file, _h5py.File):
                    filenames.append(file.filename)
                elif isinstance(file, _h5py.Group):
                    filenames.append(file.file.filename)
                    ds_names[i] = file.file.name + ds_names[i]
                else:
                    raise TypeError('if datasets are not h5py datasets, a file must be specified as h5py.File or filename(s)')
            else:
                raise TypeError('datasets should be list of strings or datasets')

        with _h5py.File(filenames[0], 'r') as h5file:
            shape = h5file[ds_names[0]].shape
            self._dtype = h5file[ds_names[0]].dtype
            ctype = _np.ctypeslib.as_ctypes_type(self._dtype)

        for d, filename in zip(ds_names[1:], filenames[1:]):
            with _h5py.File(filename, 'r') as h5file:
                if h5file[d].shape != shape:
                    raise ValueError('should have same shape')
                if _np.ctypeslib.as_ctypes_type(h5file[d].dtype) is not ctype:
                    raise TypeError('should have same dtype')

        nds = len(ds_names)
        self._nds = nds
        self._n = shape[0]
        self._lastread = _mp.Value('i', -1)
        if willread is not None:
            if not isinstance(willread, _np.ndarray) or not willread.shape[0] == shape[0] or not willread.dtype == bool:
                raise ValueError('willread must be a boolean array of same length as the datasets or None')
            if _np.all(willread):
                willread = None
            else:
                self._lastread.value = _np.argmax(willread) - 1
        shapecache = (sizecache, *shape[1:])
        self._rindices = [_mp.RawArray(_ctypes.c_int64, int(sizecache)) for i in range(nds)]
        self._indices = [_np.frombuffer(r, _np.int64) for r in self._rindices]
        for i in self._indices:
            i[:] = -5
        self._rcache = [_mp.RawArray(ctype, int(_np.prod(shapecache))) for i in range(nds)]
        self._willread = willread
        self._cache = [_np.frombuffer(r, self._dtype).reshape(shapecache) for r in self._rcache]

        self._readevents = [_mp.Event() for i in range(nds)]
        self._writeevents = [_mp.Event() for i in range(nds)]
        self._readerprocess = [
            _mp.Process(
                name=f'reader_process_{i}',
                target=self._reader,
                args=(
                    self._rindices[i],
                    self._rcache[i],
                    shapecache,
                    self._lastread,
                    self._readevents[i],
                    self._writeevents[i],
                    filenames[i],
                    ds_names[i],
                    willread,
                ),
            )
            for i in range(nds)
        ]
        for p in self._readerprocess:
            p.start()

    @staticmethod
    def _reader(rindices, rcache, shape, lastread, readevent, writeevent, filename, dsname, willread):
        """
        this function will run in separete processes to read the data into the cache
        """
        try:
            import mkl
            mkl.set_num_threads_local(1)
        except ImportError:
            pass

        if willread is not None:
            readall = False
            readindices = _np.nonzero(willread)[0]
            inextwrite = 0
        else:
            readall = True
            nextwrite = 0

        with _h5py.File(filename, 'r') as file:
            dataset = file[dsname]
            n = dataset.shape[0]
            indices = _np.frombuffer(rindices, _np.int64)
            cache = _np.frombuffer(rcache, dataset.dtype).reshape(shape)
            while True:
                if lastread.value > n:
                    return
                empty = _np.where(indices < (lastread.value - 1))[0]
                nempty = len(empty)
                if nempty == 0:
                    if not readevent.wait(0.1):
                        writeevent.set()
                    readevent.clear()
                    continue
                if readall:
                    start = max(nextwrite, lastread.value)
                    if start >= n:
                        return
                    if start + nempty > n - 1:
                        nempty = n - start
                        empty = empty[:nempty]
                    nextwrite = start + nempty
                    read = dataset[start:nextwrite, ...]
                    cache[empty, ...] = read
                    indices[empty] = _np.arange(start, nextwrite)
                    writeevent.set()
                else:
                    if readindices[inextwrite] < lastread.value:
                        inextwrite = _np.argmax(readindices > lastread.value)
                    if inextwrite + nempty > readindices.shape[0]:
                        nempty = readindices.shape[0] - inextwrite
                        empty = empty[:nempty]
                    readids = readindices[inextwrite : inextwrite + nempty]
                    read = dataset[readids, ...]
                    cache[empty, ...] = read
                    indices[empty] = readids
                    inextwrite = inextwrite + nempty
                    writeevent.set()
                    if inextwrite == readindices.shape[0]:
                        return

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._lastread.value = self._n + 1
        for ev in self._readevents:
            ev.set()
        del self._cache
        del self._rcache
        for p in self._readerprocess:
            p.join(100)
            p.kill()
        del self._readerprocess

    def __getitem__(self, i):
        innerind = tuple()
        if isinstance(i, int):
            n = i
            t = slice(None, None, None)
        elif isinstance(i, tuple):
            n, t = i[:2]
            if len(i) >= 2:
                innerind = i[2:]
            if isinstance(t, int):
                t = slice(t, t + 1, None)
            if not isinstance(n, int):
                raise IndexError('image index must be int')
        else:
            raise IndexError('must give two indices, image (int) and tile (int or slice).')
        if n < 0 or n >= self._n:
            raise IndexError(f'image index must be between 0 and {self._n - 1}')
        try:
            r = range(self._nds)[t]
        except IndexError:
            raise IndexError('tile index invalid')
        if self._lastread.value > n:
            raise IndexError('must request images monotonically increasing')
        if self._willread is not None and not self._willread[n]:
            raise IndexError(f'willread {n} was set to false, but trying to read!')
        self._lastread.value = n
        ret = []
        for ct in r:
            self._readevents[ct].set()
            while True:
                index = _np.where(self._indices[ct] == n)[0]
                if len(index) == 1:
                    ind = (index[0], *innerind)
                    ret.append(self._cache[ct][ind])
                    break
                self._writeevents[ct].wait()
                self._writeevents[ct].clear()

        return _np.squeeze(_np.asarray(ret))

    def __len__(self):
        return self._n

    @property
    def shape(self):
        return (self._n, self._nds)

    @property
    def dtype(self):
        return self._dtype


@_contextlib.contextmanager
def arrayreader(array, *args, **kwargs):
    """
    dummy context manager around an array
    """
    yield array



import numpy as _np
import cupy as _cp
from pathlib import Path as _Path


code = (_Path(__file__).parent / 'sim.cu').read_text()


def _pinned(shape, dtype):
    size = int(_np.prod(shape))
    mem = _cp.cuda.alloc_pinned_memory(size * _np.dtype(dtype).itemsize)
    ret = _np.frombuffer(mem, dtype, size).reshape(shape, order='C')
    return ret


def simulate_gen(simobject, Ndet, pixelsize, detz, k, settings="double", maximg=_np.inf, *args, **kwargs):
    """
    returns an array of simulated wavefields
    parameters:
    simobject: a simobject whose get() returns an Nx4 array with atoms in the first and (x,y,z,phase) of each atom in the last dimension
    Ndet: pixels on the detector
    pixelsize: size of one pixel in same unit as simobjects unit (usally um)
    detz: detector distance in same unit as simobjects unit (usally um)
    k: angular wavenumber
    settings: string, default: double_ff_nodist can contain
        single: use single precision
        nf:     use near field
        scale: apply 1/r intensity scaling
        secondorder: use second order in far field approximation (Fresnel)
        nofast: no fast math
        unknown options will be silently ignored.
    init: do full initialisation and asynch start of first calculation on generator creation
    maximg: generate StopIteration after maximg images.
    """

    if "single" in settings:
        intype, outtype, kernelname = _np.float32, _np.complex64, "wfkernelf"
    else:
        intype, outtype, kernelname = _np.float64, _np.complex128, "wfkerneld"
    options = []
    if "scale" not in settings:
        options += ["-Dnodist"]
    if "nf" in settings:
        options += ["-Dusenf"]
    if "secondorder" in settings:
        options += ["-Dsecondorder"]
    if "nofast" not in settings:
        options += ["--use_fast_math"]

    if _np.size(Ndet) == 1:
        Ndet = [Ndet, Ndet]

    threadsperblock = (16, 16, 1)
    blockspergrid_x = int(_np.ceil(Ndet[0] / threadsperblock[0]))
    blockspergrid_y = int(_np.ceil(Ndet[1] / threadsperblock[1]))
    blockspergrid = (blockspergrid_x, blockspergrid_y)

    module = _cp.RawModule(code=code, backend="nvcc", options=tuple(["--std=c++11", "-O3", "--restrict"] + options))
    kernel = module.get_function(kernelname)

    d_wf = _cp.zeros((Ndet[0], Ndet[1]), dtype=outtype, order='C')
    d_atoms = _cp.zeros((simobject.N, 4), intype, order='C')
    h_atoms = _pinned((simobject.N, 4), intype)

    def _gen():
        h_atoms[:, :3], h_atoms[:, 3:] = simobject.get2()
        d_atoms.set(h_atoms)
        kernel(blockspergrid, threadsperblock, (d_wf, d_atoms, float(detz), float(pixelsize), float(k), int(Ndet[0]), int(Ndet[1]), int(simobject.N)))
        count = 1
        while True:
            if count == maximg:
                yield d_wf.get()
            elif count > maximg:
                return
            else:
                h_atoms[:, :3], h_atoms[:, 3:] = simobject.get2()
                d_atoms.set(h_atoms)
                ret = d_wf.get()
                kernel(
                    blockspergrid,
                    threadsperblock,
                    (d_wf, d_atoms, float(detz), float(pixelsize), float(k), int(Ndet[0]), int(Ndet[1]), int(simobject.N)),
                )
                yield ret
            count += 1

    return _gen()


def simulate(Nimg, simobject, Ndet, pixelsize, detz, k, settings="double", verbose=True, *args, **kwargs):
    """
    returns an array of simulated wavefields
    parameters:
    Nimg: number of wavefields to simulate
    simobject: a simobject whose get() returns an Nx4 array with atoms in the first and (x,y,z,phase) of each atom in the last dimension
    Ndet: pixels on the detector
    pixelsize: size of one pixel in same unit as simobjects unit (usally um)
    detz: detector distance in same unit as simobjects unit (usally um)
    k: angular wavenumber
    settings: string, default: double_ff_nodist can contain
        single: use single precision
        nf:     use near field
        scale: apply 1/r intensity scaling
        secondorder: use second order in far field approximation (Fresnel)
        nofast: no fast math
        unknown options will be silently ignored.
    """
    if _np.size(Ndet) == 1:
        Ndet = [Ndet, Ndet]
    gen = simulate_gen(simobject, Ndet, pixelsize, detz, k, settings=settings, init=True, maximg=Nimg)
    result = _np.empty((Nimg, Ndet[0], Ndet[1]), _np.complex128)
    try:
        for i, img in enumerate(gen):
            if verbose:
                print(i, end='. ', flush=True)
            result[i, ...] = img
        return result
    except KeyboardInterrupt:
        if verbose:
            print('Interrupted')
        return result[:i]



import numba as _numba
from numba import cuda
import numpy as _np
import math
import contextlib
import warnings


@contextlib.contextmanager
def corrfunction(shape, z, maxq, xcenter=None, ycenter=None, return_gpu=False):
    """
    GPU based radial Autocorrelation with q correction

    parameters:
    shape (tuple) of inputs in pixels
    z (scalar) distance of detector in pixels
    maxq (scalar): maxmum distance
    optional:
    xcenter (scalar): position of center in x direction, defaults to shape[0]/2
    ycenter (scalar): position of center in x direction, defaults to shape[1]/2
    returns a function with signature float[:](float[:,:] image) that does the correlation
    """
    stream = None
    saveblocks = _numba.int32(16)
    try:
        if not 15 <= maxq <= 511:
            warnings.warn("maxq will be clamped between 15 and 511")
        cuda.select_device(0)
        stream = cuda.stream()
        qmax = _numba.int32(max(15, min(511, maxq)))
        Nr, Nc = int(shape[0]), int(shape[1])
        y, x = _np.meshgrid(_np.arange(Nc, dtype=_np.float64), _np.arange(Nr, dtype=_np.float64))
        xcenter = _numba.float32(xcenter or Nr / 2)
        ycenter = _numba.float32(ycenter or Nc / 2)
        x -= xcenter
        y -= ycenter
        d = _np.sqrt(x ** 2 + y ** 2 + z ** 2)
        qx, qy, qz = ((k * z / d).astype(_np.float32) for k in (x, y, z))
        with stream.auto_synchronize():
            dqx = _numba.cuda.to_device(qx, stream)
            dqy = _numba.cuda.to_device(qy, stream)
            dqz = _numba.cuda.to_device(qz, stream)
            doutput = _numba.cuda.device_array((saveblocks, qmax + 1), _np.float64, stream=stream)

        del x, y, d, qx, qy, qz
        # shared memory size is limited, don't cache if rows are to large
        # instead, we could load them blockwise..
        shareval = shape[1] <= 4096
        sharerefq = shape[1] <= 2048
        sharetq = shape[1] <= 1024

        def kernel(qx, qy, qz, val, out):
            refr = _numba.cuda.blockIdx.y
            offsetr = _numba.cuda.blockIdx.z

            saveblock = _numba.int32(_numba.cuda.threadIdx.x % saveblocks)
            tr = _numba.int32(refr + offsetr)

            if not (0 <= tr < Nr and 0 <= refr < Nr):
                return

            if sharerefq:
                refq0, refq1, refq2 = (
                    _numba.cuda.shared.array(Nc, _np.float32),
                    _numba.cuda.shared.array(Nc, _np.float32),
                    _numba.cuda.shared.array(Nc, _np.float32),
                )
            else:
                refq0, refq1, refq2 = qx[refr, :], qy[refr, :], qz[refr, :]
            if sharetq:
                tq0, tq1, tq2 = (
                    _numba.cuda.shared.array(Nc, _np.float32),
                    _numba.cuda.shared.array(Nc, _np.float32),
                    _numba.cuda.shared.array(Nc, _np.float32),
                )
            else:
                tq0, tq1, tq2 = qx[tr, :], qy[tr, :], qz[tr, :]
            if shareval:
                refv = _numba.cuda.shared.array(Nc, _np.float32)
                tv = _numba.cuda.shared.array(Nc, _np.float32)
            else:
                refv = val[refr, :]
                tv = val[tr, :]
            loadid = _numba.cuda.threadIdx.x
            _numba.cuda.syncthreads()
            if shareval or sharetq or sharerefq:
                while loadid < _numba.int32(Nc):
                    if sharerefq:
                        refq0[loadid] = qx[refr, loadid]
                        refq1[loadid] = qy[refr, loadid]
                        refq2[loadid] = qz[refr, loadid]
                    if sharetq:
                        tq0[loadid] = qx[tr, loadid]
                        tq1[loadid] = qy[tr, loadid]
                        tq2[loadid] = qz[tr, loadid]
                    if shareval:
                        refv[loadid] = val[refr, loadid]
                        tv[loadid] = val[tr, loadid]
                    loadid += _numba.cuda.blockDim.x
            _numba.cuda.syncthreads()

            offsetc = _numba.int32(_numba.cuda.threadIdx.x - qmax)
            val = _numba.float32(0)
            dqold = _numba.cuda.threadIdx.x // 2
            dq = qmax + _numba.int32(1)
            for refc in range(_numba.int32(0), _numba.int32(Nc)):
                tc = _numba.int32(refc + offsetc)
                if 0 <= tc < Nc:
                    dq = _numba.int32(round(math.sqrt((refq0[refc] - tq0[tc]) ** 2 + (refq1[refc] - tq1[tc]) ** 2 + (refq2[refc] - tq2[tc]) ** 2)))
                    if dq <= qmax:
                        if dq != dqold:
                            _numba.cuda.atomic.add(out, (saveblock, dqold), _numba.float64(val))
                            dqold = dq
                            val = _numba.float32(0)
                        val += _numba.float32(refv[refc] * tv[tc])
            if dq < qmax:
                _numba.cuda.atomic.add(out, (saveblock, dq), _numba.float64(val))

        def reduce(vals, out):
            if _numba.cuda.threadIdx.x < qmax:
                out[_numba.cuda.threadIdx.x] = vals[0, _numba.cuda.threadIdx.x]
                for i in range(1, vals.shape[0]):
                    out[_numba.cuda.threadIdx.x] += vals[i, _numba.cuda.threadIdx.x]

        def zero(vals):
            _numba.cuda.syncthreads()
            vals[_numba.cuda.blockIdx.x, _numba.cuda.threadIdx.x] = 0

        jkernel = _numba.cuda.jit("void(float32[:,:],float32[:,:],float32[:,:],float32[:,:],float64[:,:])", fastmath=True)(kernel)
        jkernel = jkernel[[(1, Nr, qmax), (int(2 * qmax + 1), 1, 1), stream]]
        jzero = _numba.cuda.jit("void(float64[:,:])")(zero)
        jzero = jzero[(doutput.shape[0]), (doutput.shape[1]), stream]
        jreduce = _numba.cuda.jit("void(float64[:,:],float64[:])")(reduce)
        jreduce = jreduce[1, doutput.shape[1], stream]

        def corr(input):
            """
            Do the correlation
            """
            if stream is None:
                raise ValueError("already closed, use within with statement")
            dvals = cuda.to_device(input.astype(_np.float32), stream)
            jzero(doutput)
            jkernel(dqx, dqy, dqz, dvals, doutput)
            out = cuda.device_array(int(qmax), stream=stream)
            jreduce(doutput, out)
            if return_gpu:
                return out
            else:
                return out.copy_to_host(stream=stream)

        yield corr
    finally:
        if stream is not None:
            stream.synchronize()
        stream = dqx = dqy = dqz = doutput = jkernel = jreduce = jzero = None


if __name__ == "__main__":
    qmax = 256
    z = 2000
    input = _np.ones((512, 512))
    with corrfunction(input.shape, z, qmax) as f:
        out = f(input)
        print(out.sum())



import numpy as _np
import numexpr as _ne


class accumulator:
    def __init__(self, maxmin=False):
        """
        Accumulate nd arrays and calculate mean and variance
        :param maxmin: Keep track of maximum and minimum
        """
        self._n = _np.array(0)
        self._mean = None
        self._nvar = None
        self._maxmin = maxmin
        self._result = None

    def __repr__(self):
        return 'accumulator[%i]' % self._n

    #  basic idea:
    #  count += 1
    #  delta = newValue - mean
    #  mean += delta / count
    #  delta2 = newValue - mean
    #  M2 += delta * delta2

    def add(self, value, count=1.0):
        self._result = None
        count = _np.array(count, value.dtype)
        if self._mean is None:
            with _np.errstate(divide='ignore', invalid='ignore'):
                self._mean = value / count
                self._mean[~_np.isfinite(self._mean)] = 0
            self._nvar = _np.zeros_like(value)
            self._n = _np.array(count, self._mean.dtype)
            if self._maxmin is True:
                self._max = _np.copy(value)
                self._min = _np.copy(value)
        else:
            tmp = _ne.evaluate('value/count - mean', local_dict={'count': count, 'value': value, 'mean': self._mean})
            with _np.errstate(divide='ignore', invalid='ignore'):
                _ne.evaluate('where(count>0,n+1,n)', local_dict={'count': count, 'n': self._n}, out=self._n)
                _ne.evaluate(
                    'where(count>0,mean+tmp/n,mean)',
                    local_dict={'value': value, 'tmp': tmp, 'mean': self._mean, 'count': count, 'n': self._n},
                    out=self._mean,
                )
                _ne.evaluate(
                    'where(count>0, nvar + tmp * (value/count - mean), nvar)',
                    local_dict={'count': count, 'value': value, 'tmp': tmp, 'mean': self._mean, 'nvar': self._nvar},
                    out=self._nvar,
                )
            if self._maxmin is True:
                if isinstance(count, _np.ndarray) and len(count.shape) > 0:
                    self._max[count > 0] = _np.maximum(self._max, value)[count > 0]
                    self._min[count > 0] = _np.minimum(self._min, value)[count > 0]
                else:
                    _np.maximum(self._max, value, out=self._max)
                    _np.minimum(self._min, value, out=self._min)

    def combine(self, other):
        if not self.shape == other.shape:
            raise ValueError('not the same shape')
        if self._maxmin:
            if not other._maxmin:
                raise ValueError('other does not have minmax set, but self has. would loose min/max information')
            self._min = _np.minimum(self._min, other._min)
            self._max = _np.maximum(self._max, other._max)

        ld = {'s_nvar': self._nvar, 's_mean': self._mean, 's_n': self._n, 'o_nvar': other._nvar, 'o_mean': other._mean, 'o_n': other._n}
        self._nvar = _ne.evaluate('s_nvar+(o_nvar + (s_mean - o_mean) ** 2 * s_n * o_n / (s_n + o_n))', local_dict=ld)
        self._mean = _ne.evaluate('(s_mean*s_n + o_mean* o_n)/(s_n + o_n)', local_dict=ld)
        self._n += other._n

    def __len__(self):
        return _np.max(self._n)

    @property
    def mean(self, copy=False):
        if self._n.size > 1 or copy:
            mean = _np.copy(self._mean)
            mean[self.n == 0] = _np.nan
            return mean
        else:
            return self._mean

    @property
    def n(self):
        return self._n

    @property
    def var(self):
        if _np.any(self._n > 0):
            with _np.errstate(all='ignore'):
                return self._nvar / self._n
        else:
            return None

    @property
    def std(self):
        if _np.any(self._n > 0):
            with _np.errstate(all='ignore'):
                return _np.sqrt(self.var)
        else:
            return None

    @property
    def max(self):
        if self._maxmin is False:
            raise ValueError('set maxmin to True when creating accumulator to keep min and max')
        return self._max

    @property
    def min(self):
        if self._maxmin is False:
            raise ValueError('set maxmin to True when creating accumulator to keep min and max')
        return self._min

    @property
    def shape(self):
        if self._mean is None:
            return tuple()
        else:
            return self._mean.shape

    @property
    def result(self):
        if self._result is None:
            self._result = _frozenaccumulatorInv(self)
        return self._result


class _frozenaccumulator:
    """
    Accumulator that doesn't accumulate anymore
    """

    def __init__(self, accum):
        self._mean = _np.copy(accum.mean)
        self._std = _np.copy(accum.std)
        self._n = _np.copy(accum.n)

    @property
    def mean(self):
        return self._mean

    @property
    def std(self):
        return self._std

    @property
    def n(self):
        return self._n

    @property
    def shape(self):
        return self._mean.shape


class _frozenaccumulatorInv(_frozenaccumulator):
    """
    Accumulator that doesn't accumulate anymore but has inverses precalculated
    """

    def __init__(self, accum):
        super().__init__(accum)
        with _np.errstate(divide='ignore', invalid='ignore'):
            self._invmean = _np.nan_to_num(1.0 / self._mean)
            self._invstd = _np.nan_to_num(1.0 / self._std)

    @property
    def invstd(self):
        return self._invstd

    @property
    def invmean(self):
        return self._invmean



import numpy as _np
import scipy.signal as _ss


def filter_std(image, size, sigma=1):
    """
    explicitly filters an nd image based on deviation by local stdev over local mean.
    return mask of points that are sigma local stdev over(under) the local mean if sigma is positiv(negativ)
    """
    import numpy as np
    import ctypes
    import scipy.ndimage as ndi
    from numba import cfunc, carray
    from numba.core.types import intc, intp, float64, voidptr, CPointer
    from scipy import LowLevelCallable

    @cfunc(intc(CPointer(float64), intp, CPointer(float64), voidptr), fastmath=True)
    def _std(values_ptr, len_values, result, data):
        values = carray(values_ptr, (len_values,), dtype=float64)
        accumx = 0
        accumx2 = 0
        sigma = carray(data, (1,), dtype=float64)[0]
        for x in values:
            accumx += x
            accumx2 += x * x
        mean = accumx / len_values
        std = np.sqrt((accumx2 / len_values) - mean ** 2)
        result[0] = sigma * std + mean
        return 1

    csigma = ctypes.c_double(sigma)
    ptr = ctypes.cast(ctypes.pointer(csigma), ctypes.c_void_p)
    res = image > ndi.generic_filter(image, LowLevelCallable(_std.ctypes, ptr), size)
    if sigma < 0:
        res = ~res
    return res


def std(image, size):
    """
    a moving std calculation on an nd image
    """
    import numpy as np
    import ctypes
    import scipy.ndimage as ndi
    from numba import cfunc, carray
    from numba.core.types import intc, intp, float64, voidptr, CPointer
    from scipy import LowLevelCallable

    @cfunc(intc(CPointer(float64), intp, CPointer(float64), voidptr), fastmath=True)
    def _std(values_ptr, len_values, result):
        values = carray(values_ptr, (len_values,), dtype=float64)
        accumx = 0
        accumx2 = 0
        for x in values:
            accumx += x
            accumx2 += x * x
        mean = accumx / len_values
        std = np.sqrt((accumx2 / len_values) - mean ** 2)
        result[0] = std
        return 1

    res = ndi.generic_filter(image, LowLevelCallable(_std.ctypes), size)
    return res


def fftfilter_mean(image, size, norm=False):
    strel = _np.ones(image.ndim * [size]) / size ** image.ndim
    res = _ss.fftconvolve(image, strel, mode='same')
    if norm:
        res /= _ss.fftconvolve(_np.ones_like(image), strel, mode='same')
    return res


def fftfilter_std(image, size, sigma):
    norm = fftfilter_mean(_np.ones_like(image), size)
    mean = fftfilter_mean(image, size) / norm
    sqmean = fftfilter_mean(image * image, size) / norm
    std = _np.nan_to_num(_np.sqrt(sqmean - mean * mean))
    res = _np.abs(image - mean) > sigma * std
    res[_np.abs(sigma * std) < 1e-15] = 0
    return res



__all__ = ['hitcor', 'hitcorrad', 'cpucor', 'cpucorrad', 'cucor', 'cucorrad', 'cpusimple', 'cusimple', 'ft', 'singleshotnorm']
try:
    import mkl as _mkl

    _vml_threads = _mkl.domain_get_max_threads('vml')
except ImportError:
    _mkl = None

import numba.cuda as _nbcuda
import warnings as _w

_w.filterwarnings("ignore", message="numpy.dtype size changed")
_w.filterwarnings("ignore", message="numpy.ufunc size changed")


from . import hitcor, hitcorrad, cpucor, cpucorrad, cpusimple, common, singleshotnorm  # noqa

# mkl fft
try:
    from . import autocorrelate3  # noqa

    _local = False
except ImportError as e:
    _w.warn('trying local compilation' + str(e))
    _local = True

    try:
        import pyximport as _pyx
        from sys import prefix as _prefix
        from os.path import join as _join
        from numpy.distutils.system_info import get_info as _getinfo, default_include_dirs as _defaultincludedirs
        from numpy import get_include as _np_get_include

        _incs = _defaultincludedirs
        _incs.append(_np_get_include())
        _incs.append(_join(_prefix, 'include'))
        _mklinc = _getinfo('mkl').get('include_dirs')
        if _mklinc:
            _incs.extend(_mklinc)
        _pyx.install(setup_args={'include_dirs': _incs}, language_level=2)
    except Exception as e:
        _w.warn("no mkl autocorrelation")

from . import ft  # noqa

# cuda
_cuda = _nbcuda.is_available()
if _cuda:
    try:
        from . import cucor, cucorrad, cusimple

        simple = cusimple
        qcor = cucor
        qcorrad = cucorrad

    except Exception as e:
        print(e)
        _cuda = False
if not _cuda:
    _w.warn("no cuda available")
    simple = cpusimple
    qcor = cpucor
    qcorrad = cpucorrad

if _mkl is not None:
    _mkl.domain_set_num_threads(_vml_threads, 'vml')  # numexpr messes with vml thread number



import numba as _numba
import numpy as _np


@_numba.njit(nogil=True)
def _getidx(p, pmax, idmax):
    """
    gives ids (up to idxmax) for job p of pmax jobs to do. will return ids in blocks and pair low ids with high ids for the same job
    """
    idsperP = idmax // (2 * pmax)
    n = 2 * idsperP
    missing = idmax - (idsperP * (2 * pmax))
    if p < missing:
        if p < missing - pmax:
            n = n + 2
            out = _np.empty(n, dtype=_numba.int64)
            out[-1] = idmax - p - pmax - 1
            out[-2] = idmax - p - 1
        else:
            n = n + 1
            out = _np.empty(n, dtype=_numba.int64)
            out[-1] = idmax - p - 1
    else:
        out = _np.empty(n, dtype=_numba.int64)
    for (m, x) in enumerate(range(p * idsperP, (p + 1) * idsperP)):
        out[m] = x
    for (k, x) in enumerate(range(((2 * pmax - 1) - p) * idsperP, ((2 * pmax) - p) * idsperP)):
        out[m + k + 1] = x
    return out


def getidx_ordered(p, pmax, idmax):
    """
    gives ids (up to idxmax) for job p of pmax jobs to do.
    """
    return [i for i in range(p, idmax, pmax)]



__all__ = ['cpu', 'simobj', 'cuda', 'common', 'time', 'cutime', 'simple']

try:
    import mkl as _mkl

    _vml_threads = _mkl.domain_get_max_threads('vml')
except ImportError:
    _mkl = None

from . import cpu, simobj, time, simple
from .common import *

try:
    from . import cuda

    auto = cuda
except ImportError as _e:
    if any(x in str(_e.args[0]).lower() for x in ["cuda", "libcu", "cupy"]):
        import warnings as _w

        _w.warn('cuda error. cuda time dependent simulation not imported. is cuda available and paths set?')
        auto = cpu
    else:
        print(_e)
except Exception as _e:
    print('Import exception on cuda:')
    print(_e)
try:
    from . import cutime

    autotime = cutime
except (AttributeError, ImportError) as _e:
    if any(x in str(_e.args[0]).lower() for x in ["cuda", "libcu", "cupy"]):
        import warnings as _w

        _w.warn('cuda error. cuda time dependent simulation not imported. is cuda available and paths set?')
        autotime = time
    else:
        print(_e)
except Exception as _e:
    print('Import exception on cutime:')
    print(_e)

if _mkl is not None:
    _mkl.domain_set_num_threads(_vml_threads, 'vml')  # numexpr messes with vml thread number



import cupy as _cp
import numpy as _np
from pathlib import Path as _Path


code = (_Path(__file__).parent / 'simtime.cu').read_text()


def simulate(simobject, Ndet, pixelsize, detz, k, c, tau, pulsewidth, settings='mixed', threads=None):
    """
    Time dependent simulation with decaying amplitudes.
    simobject: simobject to use for simulation (in lengthunit)
    pixelsize: pixelsize (in lengthunit)
    detz: Detector-sample distance
    k: angular wave number (in 1/lengthunit)
    c: speed of light in (lengthunit/timeunit)
    tau: decay time (in timeunit)
    pulsewidth: FWHM of gaussian exciation pulse (in timeunit)
    settings: string, can contain
        double,single,mixed - precision
        nf - for nearfield form
        scale - do 1/r intensity scaling
    first call with new settings might recompile and take a few seconds
    """

    if 'double' in settings:
        nametmp, namesim = "tempsized", "simulated"
        ttype, atype, inouttype = _np.float64, _np.complex128, _np.float64
    elif 'single' in settings:
        nametmp, namesim = "tempsizef", "simulatef"
        ttype, atype, inouttype = _np.float32, _np.complex64, _np.float32
    else:  # mixed
        nametmp, namesim = "tempsizedf", "simulatedf"
        ttype, atype, inouttype = _np.float32, _np.complex128, _np.float64
    options = []
    if 'nf' in settings:
        options += ['-Dusenf']
    if 'scale' not in settings:
        options += ['-Dnodist']
    options += ['-dc', '--std=c++11', '--expt-relaxed-constexpr', '-O3', '--use_fast_math']

    module = _cp.RawModule(code=code, backend='nvcc', options=tuple(options))
    ftempsize = module.get_function(nametmp)
    fsimulate = module.get_function(namesim)
    N = simobject.N
    if threads is None:
        if N > 1e8:
            threads = 1
        elif N > 1e7:
            threads = 2
        elif N > 1e6:
            threads = 4
        else:
            threads = 8

    if _np.size(Ndet) == 1:
        Ndet = [Ndet, Ndet]

    det = _cp.array(
        _np.array(_np.meshgrid(pixelsize * (_np.arange(Ndet[0]) - (Ndet[0] / 2)), pixelsize * (_np.arange(Ndet[1]) - (Ndet[1] / 2)), detz)).T.reshape(
            -1, 3
        ),
        inouttype,
    )
    pdet = _cp.array([i.data.ptr for i in det], _np.uint64)

    tmpout = _cp.zeros(1, _np.int64)
    ftempsize(grid=(1,), block=(1,), args=(N, tmpout))
    tempsize = int(tmpout.get()[0])
    temp = [_cp.zeros(tempsize // 8 + 1, _np.int64) for i in range(threads)]
    ptemp = _cp.array([i.data.ptr for i in temp], _np.uint64)

    t = [_cp.zeros(N, ttype) for i in range(2 * threads)]
    a = [_cp.zeros(N, atype) for i in range(2 * threads)]
    pt = _cp.array([i.data.ptr for i in t], _np.uint64)
    pa = _cp.array([i.data.ptr for i in a], _np.uint64)
    data = _cp.array(simobject.get(), inouttype, order='C')
    times = _cp.array(_np.random.randn(simobject.N) * (pulsewidth / 2.35), inouttype)
    # _cp.random.seed(_np.random.randint(2**64-1,dtype=_np.uint64))
    # times = (_cp.random.randn(N,dtype=_np.float64)*(pulsewidth/2.35)).astype(inouttype) #cave: this uses different seed as numpy
    output = _cp.zeros(len(det), inouttype)
    poutput = _cp.array([i.data.ptr for i in output], _np.uint64)
    _cp.cuda.get_current_stream().synchronize()
    for start in range(0, len(pdet), threads):
        cthreads = int(min(len(pdet) - start, threads))
        end = start + cthreads
        fsimulate(
            grid=(1,),
            block=(cthreads,),
            args=(data, times, pdet[start:end], float(tau), float(c), float(k), int(N), int(cthreads), pt, pa, tempsize, ptemp, poutput[start:end],),
        )
    _cp.cuda.get_current_stream().synchronize()
    return output.get().reshape(Ndet)



import numpy as _np
from ..util import fastlen as _fastlen

# import numexpr as _ne
# from ..util import abs2
# def corr(input):
#     ret = _np.zeros([2 * s for s in input.shape])
#     ret[:input.shape[0], :input.shape[1]] = input
#     ret = _np.fft.fftshift(_np.fft.irfftn(abs2(_np.fft.rfftn(ret))))
#     return ret


def corr(input, axes=(-1, -2), norm=False, fftfunctions=(_np.fft.rfftn, _np.fft.irfftn), **kwargs):
    """
    simple autocorrelation of input along axes (default: last two)
    axes: axes to correlate along, defaults to last two
    norm: do normalisation along non correlation axes and normalise for pair count
    """
    fft, ifft = fftfunctions
    axes = sorted([input.ndim + a if a < 0 else a for a in axes])
    fftshape = [_fastlen(2 * input.shape[ax]) for ax in axes]
    if norm:
        input = input * (1 / input.mean(axis=[i for i in range(input.ndim) if i not in axes] or None))
    ret = fft(input, fftshape, axes=axes)
    ret = _np.abs(ret) ** 2
    # _ne.evaluate('(ret*conj(ret))', out=ret, casting='same_kind')
    ret = ifft(ret, axes=axes)
    ret = _np.fft.fftshift(ret, axes=axes)[
        tuple((Ellipsis, *(slice(ps // 2 - input.shape[ax], ps // 2 + input.shape[ax]) for ax, ps in zip(axes, fftshape))))
    ]

    if norm:
        n = corr(_np.ones(tuple(input.shape[ax] for ax in axes)), fftfunctions=fftfunctions)
        ret /= n
        ret[(...,) + (n < 0.9).nonzero()] = _np.nan
    return ret



from IPython.core.magic import Magics, magics_class, line_magic, cell_magic
import argparse
import shlex
import os
import time
import subprocess
import getpass
import IPython.display

"""
IPython magic to submit jobs
"""


@magics_class
class submit_magics(Magics):
    defaultline = ''

    @classmethod
    def _run(cls, defaultcmd, key, mapping, vs, line, cell, defaultline=''):
        parser = argparse.ArgumentParser(description='Submit a job', prog=defaultcmd)
        parser.add_argument('-c', dest='c', default='4', type=str, help='n cpu per task (4)')
        parser.add_argument('-m', dest='m', default='4G', type=str, help='ram (4G)')
        parser.add_argument('-t', dest='t', default='06:00:00', type=str, help='time (6h)')
        parser.add_argument('-n', dest='n', default='submitmagic', type=str, help='name')
        parser.add_argument('-o', dest='o', default='~/joblog_%j.log', type=lambda x: os.path.expanduser(x), help="output (~/joblog_%%j.log)")
        parser.add_argument('-g', dest='g', default=None, type=str, help='gpus (None)')
        parser.add_argument('-a', dest='a', default=None, type=str, help='array')
        parser.add_argument('--interpreter', default='python', type=str, help='interpreter (python)')
        parser.add_argument('--batch_cmd', metavar='cmd', type=str, default=defaultcmd, nargs='*', help='batch command')
        parser.add_argument('--dryrun', action='store_true', help="don't submit")
        parser.add_argument('--debug', action='store_true', help="print debug info")
        parser.add_argument('--no-replace', dest='replace', action='store_false', help="don't replace scheduler variables with values")
        parser.epilog = f'defaults can be set with %submitdefault line magic. {"currently set:" if defaultline else ""} {defaultline}'

        try:
            args, u = parser.parse_known_args(shlex.split(' '.join([line, defaultline])))
        except SystemExit:
            return

        unknown = []
        for el in u:
            if el[0] == '-':
                unknown.append(el)
            elif len(unknown) > 0:
                unknown[-1] += ' ' + el
        if len(vs) and args.replace:
            prepcmd = 'sed ' + ' '.join([f'-e "s^{{{{{v}}}}}^${v}^g"' for v in vs])
        else:
            prepcmd = 'cat'
        cmd = '\n'.join(
            ['#!/bin/sh']
            + [f'{key} {mapping[k]}={v}' for k, v in vars(args).items() if (v is not None and k in mapping)]
            + [f'{key} {v}' for v in unknown]
            + (['', 'hostname', 'env', 'pwd', 'mount'] if args.debug else [])
            + [
                """
_term() {
    echo "Got SIGTERM signal!"
    kill -TERM "$pid"
}
_usr1() {
    echo "Got USR1 signal!"
    kill -USR1 "$pid"
}
trap _term TERM
trap _usr1 USR1
            """
            ]
            + [f'{prepcmd} << "EOF_PYTHONFILE" | {args.interpreter} &', '', cell, '', 'EOF_PYTHONFILE']
            + ['export pid=$!']
            + (['echo "$pid"'] if args.debug else [])
            + ['wait']
        )
        bcmd = [inner for outer in [shlex.split(i) for i in args.batch_cmd] for inner in outer]
        if args.debug or args.dryrun:
            print(args.batch_cmd, bcmd)
            print(cmd)
        if not args.dryrun:
            try:
                p = subprocess.run(bcmd, input=cmd.encode('utf-8'), capture_output=True)
            except FileNotFoundError:
                return 'Submit Command not found'
            if p.returncode or args.debug:
                print(p.stderr.decode('utf-8'))
                print(p.stdout.decode('utf-8'))
            return p.returncode

    @line_magic
    def submitdefault(self, line):
        if line:
            self.defaultline = self.defaultline + ' ' + line
        else:
            print(
                f"""
usage: %submitdefault -argument val
use reload_ext to reset.
currently set:
{self.defaultline}"""
            )

    @cell_magic
    def slurm(self, line, cell=None):
        """
        usage: %%slurm [-h] [-c C] [-m M] [-t T] [-n N] [-o O] [-g G] [-a A] [-i I]
        [--dryrun] [--debug] [--no-replace]
        [sbatch_cmd [sbatch_cmd ...]]

        positional arguments:
        sbatch_cmd    batch command, e.g sbatch (default), ssh loginserver sbatch, etc.

        optional arguments:
        -h, --help    show a help message
        -c C          n cpu per task (4)
        -m M          ram (4G)
        -t T          time (6h)
        -n N          name
        -o O          output (~/joblog_%j.log)
        -g G          gpus (None)
        -a A          array
        -i I          interpreter (python)
        --dryrun      don't submit
        --debug       print debug info
        --no-replace  don't replace scheduler variables with values
        """
        defaultcmd = 'sbatch'
        mapping = {
            'c': '--cpus-per-task',
            'm': '--mem',
            't': '--time',
            'n': '--job-name',
            'o': '--output',
            'g': '--gpus',
            'a': '--array',
        }
        key = '#SBATCH'
        vs = [
            'SLURM_JOBID',
            'SLURM_ARRAY_TASK_MAX',
            'SLURM_ARRAY_TASK_MIN',
            'SLURM_ARRAY_TASK_COUNT',
            'SLURM_ARRAY_TASK_ID',
            'SLURM_ARRAY_TASK_STEP',
            'SLURM_ARRAY_JOB_ID',
            'SLURM_NODELIST',
            'SLURM_JOB_NAME',
            'SLURM_GPUS',
            'SLURM_JOB_CPUS_PER_NODE',
            'SLURM_MEM_PER_NODE',
        ]
        return submit_magics._run(defaultcmd, key, mapping, vs, line, cell, self.defaultline)

    @cell_magic
    def pbs(self, line, cell=None):
        raise NotImplementedError('pbs not implmented')

    @cell_magic
    def lsf(self, line, cell=None):
        raise NotImplementedError('lsf not implmented')

    @line_magic
    def queue(self, line=None, display=None):
        cmd = shlex.split(line) or [
            '/opt/slurm/bin/squeue',
            '-u',
            getpass.getuser(),
            '-O',
            'JobArrayID:80,JobID:80,name:80,state:80,ReasonList:80,MINCPUS:80,tres-per-job:80,MinMemory:80,SubmitTime:80,TimeUsed:80,TimeLeft:80',
        ]
        display = display or IPython.display.display({'text/plain': 'getting data..'}, raw=True, display_id=True)
        p = subprocess.run(cmd, capture_output=True, timeout=5)
        if not p.returncode:
            output = p.stdout.decode('ascii').split("\n")
            rows = [f'<th>{"</th><th>".join(output[0].split())}</th>'] + [f'<td>{"</td><td>".join(line.split())}</td>' for line in output[1:-1]]
            html = "<table><tr>" + '</tr>\n<tr>'.join(rows) + '</tr><table>'
            display.update({'text/html': html + f'<i>Last Updated at {time.asctime(time.localtime())}</i>'}, raw=True)
            time.sleep(1)
        else:
            return p.returncode

    @line_magic
    def monitor(self, line=None):
        display = IPython.display.display({'text/plain': 'getting data..'}, raw=True, display_id=True)
        try:
            while True:
                if self.queue(line, display):
                    break
        except KeyboardInterrupt:
            IPython.display.display({'text/plain': 'Stopped'}, raw=True)


def load_ipython_extension(ipython):
    ipython.register_magics(submit_magics)



import numpy as _np
import numba as _numba
from .common import _getidx
import math as _math


def corr(input, z, qmax=None):
    """
    radial profile of correlation of hits
    for one NxN array or nx(NxN) arrays
    """
    if input.ndim == 2:
        return _pradcorr(input, z, qmax)
    elif input.ndim == 3:
        return _np.sum(_pradcorrs(input, z, qmax), axis=0)
    else:
        raise TypeError


@_numba.njit(fastmath=True)
def _radcorr(input, z, qmax=None):
    """
    radial profile of correlation
    for one NxN array
    """
    Nx, Ny = input.shape
    xi, yi = _np.where(input > 0)
    Nhits = len(xi)
    qmax = qmax or int(_np.ceil(2 * max(input.shape[-2:])))
    tmp = _np.zeros(qmax, dtype=_np.float64)
    x = xi.astype(_numba.float64) - Nx / 2.0
    y = yi.astype(_numba.float64) - Ny / 2.0
    d = _np.sqrt(x ** 2 + y ** 2 + z ** 2)
    kx = (x / d) * z
    ky = (y / d) * z
    kz = (z / d) * z

    for n in range(Nhits):
        for m in range(n):
            qz = kz[n] - kz[m]
            qx = kx[n] - kx[m]
            qy = ky[n] - ky[m]
            q = int(round(_math.sqrt(qx ** 2 + qy ** 2 + qz ** 2)))
            if 0 <= q < qmax:
                tmp[q] += input[xi[n], yi[n]] * input[xi[m], yi[m]]
    return tmp


@_numba.njit(parallel=True, cache=False, nogil=True, fastmath=True)
def _pradcorr(input, z, qmax=None):
    """
    radial profile of correlation
    for one NxN array (parallel version)
    """
    pmax = 16
    Nx, Ny = input.shape
    xi, yi = _np.where(input)
    x = xi.astype(_numba.float64) - Nx / 2.0
    y = yi.astype(_numba.float64) - Ny / 2.0
    Nhits = len(xi)
    qmax = qmax or int(_np.ceil(2 * max(input.shape[-2:])))
    tmp = _np.zeros((pmax, qmax), dtype=_numba.float64)
    d = _np.sqrt(x ** 2 + y ** 2 + z ** 2)
    kx = (x / d) * z
    ky = (y / d) * z
    kz = (z / d) * z

    for p in _numba.prange(pmax):
        idx = _getidx(p, pmax, Nhits)
        for n in idx:
            for m in range(n):
                qz = kz[n] - kz[m]
                qx = kx[n] - kx[m]
                qy = ky[n] - ky[m]
                q = int(round(_math.sqrt(qx ** 2 + qy ** 2 + qz ** 2)))
                if 0 <= q < qmax:
                    tmp[p, q] += input[xi[n], yi[n]] * input[xi[m], yi[m]]
    out = _np.zeros(qmax)
    for n in range(pmax):
        out += tmp[n, ...]
    return out


@_numba.njit(parallel=True)
def _pradcorrs(input, z, qmax=None):
    """
    radial profile of correlation
    for n NxN arrays, parallel over n
    """
    qmax = qmax or int(_np.ceil(2 * max(input.shape[-2:])))
    out = _np.zeros((input.shape[0], qmax), dtype=_numba.float64)
    print(out.shape)
    for n in _numba.prange(input.shape[0]):
        print('start', n)
        out[n] = _radcorr(input[n, ...], z, qmax)
        print('end', n)
    return out



import numba as _numba
import numpy as _np
import numexpr as _ne


@_numba.njit(parallel=True)
def _decaysum(a, t, tau):
    x = _np.zeros_like(a)
    for j in _numba.prange(x.shape[0]):
        x[j, 0] = a[j, 0]
        for i in range(1, x.shape[1]):
            decay = _np.exp(-(t[j, i] - t[j, i - 1]) / tau)
            x[j, i] = a[j, i] + decay * x[j, i - 1]
    return x


def _ab2(x):
    return x.imag ** 2 + x.real ** 2


def _integral(amp, t0, tau):
    """
    integrates the exponential decay of |x|^2 with supports t0 and decay tau
    """

    idx = _np.argsort(t0, axis=-1)
    t0s = _np.atleast_2d(t0)
    t0s = t0s[_np.arange(t0s.shape[0])[:, None], idx]
    amps = _np.atleast_2d(amp)
    amps = amps[_np.arange(amps.shape[0])[:, None], idx]
    i = _ab2(_decaysum(amps, t0s, tau))
    td = _np.diff(t0s, axis=-1)
    return _np.sum(-tau / 2 * i[:, :-1] * _np.expm1(-2 * td / tau), axis=-1) + tau / 2 * i[:, -1]


def simulate(simobject, Ndet, pixelsize, detz, k, c, tau, pulsewidth, settings=''):
    """
    Time dependent simulation with decaying amplitudes (cpu version).
    simobject: simobject to use for simulation (in lengthunit)
    pixelsize: pixelsize (in lengthunit)
    detz: Detector-sample distance
    k: angular wave number (in 1/lengthunit)
    c: speed of light in (lengthunit/timeunit)
    tau: decay time (in timeunit)
    pulsewidth: FWHM of gaussian exciation pulse (in timeunit)
    settings: string, can contain
        scale - do 1/r intensity scaling
    """

    if _np.size(Ndet) == 1:
        Ndet = [Ndet, Ndet]
    if 'scale' in settings:
        eq = 'exp(-1j*(s*k-phases))/d'
    else:
        eq = 'exp(-1j*(s*k-phases))'
    n = _np.prod(Ndet)
    # do highest power of two of n <=4 pixels at once. tradeoff between memory allocations and call overhead.
    blocksize = min(4, (n & (~(n - 1))))

    dets = _np.array(
        _np.meshgrid(
            pixelsize * (_np.arange(Ndet[0]) - (Ndet[0] / 2)),
            pixelsize * (_np.arange(Ndet[1]) - (Ndet[1] / 2)),
            detz
        )
    ).T
    res = _np.zeros(Ndet[0] * Ndet[1]).reshape(-1, blocksize)
    data = simobject.get()
    times = _np.random.randn(simobject.N) * (pulsewidth / 2.35)

    for j, det in enumerate(dets.reshape(-1, blocksize, 3)):
        d = _np.linalg.norm(det, axis=-1)[:, None]
        q = (det / d)
        q[..., -1] -= 1
        phases = data[:, 3]
        s = _np.inner(q, data[:, :3])  # path difference
        e = _ne.evaluate(eq)  # complex e field
        t = -s / c + times.T  # arrival time
        res[j] = _integral(e, t, tau)
    res = res.reshape(Ndet[0], Ndet[1])
    return res



import numpy as _np
import h5py


def appenddata(file, key, data, chunks=None, compression='lzf'):
    data = _np.atleast_1d(_np.array(data))
    if _np.array(data).dtype.kind == 'U':
        data = data.astype(h5py.string_dtype(encoding='ascii'))

    if isinstance(file, h5py.Dataset):
        if not (key is None or key == '/'):
            raise KeyError('If file is a dataset, key must be /')
        else:
            ds = file
    else:
        if key not in file.keys():
            file.create_dataset(
                key,
                chunks=chunks or data.shape,
                compression=compression,
                shuffle=True if compression else None,
                data=data,
                maxshape=(None, *data.shape[1:]),
            )
            return
        else:
            ds = file[key]
    ds.resize((ds.shape[0] + data.shape[0]), axis=0)
    ds[-data.shape[0] :] = data


def overwritedata(file, key, data, chunks=None, compression='lzf'):
    if isinstance(file, h5py.Dataset):
        raise TypeError('file should be a file or group, not a dataset')
    data = _np.atleast_1d(_np.array(data))
    if _np.array(data).dtype.kind == 'U':
        data = data.astype(h5py.string_dtype(encoding='ascii'))
    if key in file.keys():
        del file[key]
    file.create_dataset(
        key, chunks=chunks or data.shape, compression=compression, shuffle=True if compression else None, data=data, maxshape=(None, *data.shape[1:])
    )


def shrink(file, key, n):
    file[key].resize(file[key].shape[0] - n, axis=0)


def copymasked(src, dst, mask):
    def func(name, obj):
        if isinstance(obj, h5py._hl.dataset.Dataset):
            dst[name] = obj[mask, ...]

    if isinstance(src, h5py._hl.dataset.Dataset):
        name = src.name.split('/')[-1]
        dst[name] = src[mask]
    elif isinstance(src, h5py._hl.group.Group) or isinstance(h5py._hl.files.File):
        src.visititems(func)
    else:
        raise TypeError


def chunkediter(dataset, sel=slice(None), readsize=16, outsize=1):
    r = range(*sel.indices(len(dataset)))
    for i in range(0, len(r), readsize):
        ids = r[i : i + readsize]
        tmp = _np.array(dataset[ids.start : ids.stop : ids.step])
        for j in range(0, len(tmp), outsize):
            yield tmp[j : j + outsize][0] if outsize == 1 else tmp[j : j + outsize]



import numpy as _np
import numba as _numba


def get_kernel(Natoms, Ndet, pixelsize, detz, k, nodist=True, nf=False):
    """
    returns a cpu implementation of the wavefield, used internally
    Natoms: Number of atoms
    Ndet: detector pixels
    pixelsize: detector pixelsize
    detz: detector distance
    k: angular wavenumber
    returns a function with signature complex64(:,:)(out complex64(:,:), atompositionsandphases float64[:,4])
        that will write the wavefield into out
    """
    maxx, maxy = int(Ndet[0]), int(Ndet[1])
    k = float(k)
    pixelsize = float(pixelsize)
    detz = float(detz)
    Natoms = int(Natoms)

    @_numba.njit(inline='always', fastmath=True)
    def nfphase(atom, detx, dety):
        dist = _np.sqrt((detx - atom[0]) ** 2 + (dety - atom[1]) ** 2 + (detz - atom[2]) ** 2)
        phase = (dist - detz + atom[2]) * k + atom[3]
        if nodist:
            px = _np.cos(phase)
            py = _np.sin(phase)
        else:
            rdist = 1 / dist
            px = rdist * _np.cos(phase)
            py = rdist * _np.sin(phase)
        return px, py

    @_numba.njit(inline='always', fastmath=True)
    def ffphase(atom, qx, qy, qz, rdist):
        phase = -(atom[0] * qx + atom[1] * qy + atom[2] * qz) + atom[3]
        if nodist:
            px = _np.cos(phase)
            py = _np.sin(phase)
        else:
            px = rdist * _np.cos(phase)
            py = rdist * _np.sin(phase)
        return px, py

    @_numba.njit('complex128[:, ::1],float64[:, ::1]', parallel=True, fastmath={'contract', 'arcp', 'ninf', 'nnan'})
    def kernel(ret, atom):
        for x in _numba.prange(maxx):
            for y in range(maxy):
                detx = (x - maxx / 2) * pixelsize
                dety = (y - maxy / 2) * pixelsize
                accumx = 0.0
                accumy = 0.0
                cx = 0.0
                cy = 0.0
                if not nf:
                    rdist = 1 / _np.linalg.norm(_np.array([detx, dety, detz]))
                    qz = k * (detz * rdist - 1)
                    qx = k * (detx * rdist)
                    qy = k * (dety * rdist)

                for i in range(Natoms):
                    if nf:
                        px, py = nfphase(atom[i, :], detx, dety)
                    else:
                        px, py = ffphase(atom[i, :], qx, qy, qz, rdist)

                    # kahan summation
                    yx = px - cx
                    yy = py - cy
                    tx = accumx + yx
                    ty = accumy + yy
                    cx = (tx - accumx) - yx
                    cy = (ty - accumy) - yy
                    accumx = tx
                    accumy = ty

                ret[x, y] = accumx + 1j * accumy

    return kernel


def simulate(Nimg, simobject, Ndet, pixelsize, detz, k, settings='', verbose=False, *args, **kwargs):
    """
    returns an array of simulated wavefields
    parameters:
    Nimg: number of wavefields to simulate
    simobject: a simobject whose get() returns an Nx4 array with atoms in the first
        and (x,y,z,phase) of each atom in the last dimension
    Ndet: pixels on the detector
    pixelsize: size of one pixel in same unit as simobjects unit (usually um)
    detz: detector distance in same unit as simobjects unit (usually um)
    k: angular wavenumber
    settings: string
        if it contains 'scale', 1/r  scaling is performed
        if it contains 'nf', no far field approximation is made
    """

    if _np.size(Ndet) == 1:
        Ndet = [Ndet, Ndet]
    result = _np.empty((Nimg, Ndet[0], Ndet[1]), dtype=_np.complex128)
    gen = simulate_gen(simobject, Ndet, pixelsize, detz, k, settings)
    for n in range(0, Nimg):
        if verbose:
            print(n, end='', flush=True)
        result[n] = next(gen)
        if verbose:
            print('. ', end='', flush=True)
    return result


def simulate_gen(simobject, Ndet, pixelsize, detz, k, settings='', *args, **kwargs):
    """
    returns a generator that yields simulated wavefields
    parameters:
    simobject: a simobject whose get() returns an Nx4 array with atoms in the first
        and (x,y,z,phase) of each atom in the last dimension
    Ndet: pixels on the detector
    pixelsize: size of one pixel in same unit as simobjects unit (usually um)
    detz: detector distance in same unit as simobjects unit (usually um)
    k: angular wavenumber
    settings: string
        if it contains 'scale', 1/r  scaling is performed
        if it contains 'nf', no far field approximation is made
    """

    nodist = 'scale' not in settings
    nf = 'nf' in settings

    if _np.size(Ndet) == 1:
        Ndet = [Ndet, Ndet]

    result = _np.empty((Ndet[0], Ndet[1]), dtype=_np.complex128)
    f = get_kernel(simobject.N, Ndet, pixelsize, detz, k, nodist=nodist, nf=nf)
    while True:
        atoms = simobject.get()
        f(result, atoms)
        yield result



import numba as _numba
from numba import cuda
import numpy as _np
import math
import contextlib
import warnings


@contextlib.contextmanager
def corrfunction(shape, z, maxq, xcenter=None, ycenter=None):
    """
    GPU based Autocorrelation with q correction

    parameters:
    shape (tuple) of inputs in pixels
    z (scalar) distance of detector in pixels
    maxq (scalar): maxmum distance
    optional:
    xcenter (scalar): position of center in x direction, defaults to shape[0]/2
    ycenter (scalar): position of center in x direction, defaults to shape[1]/2
    returns a function with signature float[:,:,:](float[:,:] image) that does the correlation
    and returns val(dqz,dqy,dqx)=sum_qx,qy image(qx,qy)*image(qx+dqx,qy+dqy)
    """

    stream = None
    dzd = None
    dvals = None
    doutput = None
    jkernel = None

    try:
        cuda.select_device(0)
        stream = cuda.stream()
        Nr, Nc = int(shape[0]), int(shape[1])
        if not 32 <= maxq <= 1024:
            warnings.warn("maxq will be clamped between 32 and 1024")
        qmax = _numba.int32(max(32, min(1024, maxq, Nr, Nc)))
        y, x = _np.meshgrid(_np.arange(Nc, dtype=_np.float64), _np.arange(Nr, dtype=_np.float64))
        xcenter = _numba.float32(xcenter or Nr / 2)
        ycenter = _numba.float32(ycenter or Nc / 2)

        x -= xcenter
        y -= ycenter
        d = _np.sqrt(x ** 2 + y ** 2 + z ** 2)
        zd = _np.array((z / d), _np.float32, order="C")
        maxdqz = _numba.int32(math.ceil(z * (_np.max(zd) - _np.min(zd)))) | 1
        cz = _numba.float32(z)

        with stream.auto_synchronize():
            dzd = _numba.cuda.to_device(zd, stream)
            doutput = _numba.cuda.device_array((2 * maxdqz, 2 * qmax, 2 * qmax), _np.float32, stream=stream)

        del zd, x, y, d

        use_shared = Nc <= 1024

        def kernel(zd, val, out):
            refr = _numba.int32(_numba.cuda.blockIdx.y)
            offsetr = _numba.int32(_numba.cuda.blockIdx.z)
            tr = _numba.int32(refr + offsetr)
            if not (0 <= tr < Nr and 0 <= refr < Nr):
                return
            # offsetc=numba.cuda.threadIdx.x-qmax
            if use_shared:
                refzd = _numba.cuda.shared.array(Nc, _numba.float32)
                refv = _numba.cuda.shared.array(Nc, _numba.float32)
                tzd = _numba.cuda.shared.array(Nc, _numba.float32)
                tv = _numba.cuda.shared.array(Nc, _numba.float32)

                loadid = _numba.int32(_numba.cuda.threadIdx.x)
                _numba.cuda.syncthreads()
                while loadid < Nc:
                    refzd[loadid] = zd[refr, loadid]
                    refv[loadid] = val[refr, loadid]
                    tzd[loadid] = zd[tr, loadid]
                    tv[loadid] = val[tr, loadid]
                    loadid += _numba.cuda.blockDim.x
            else:
                refzd = zd[refr, :]
                refv = val[refr, :]
                tzd = zd[tr, :]
                tv = val[tr, :]

            _numba.cuda.syncthreads()
            tr2 = _numba.float32(tr) - xcenter
            refr2 = _numba.float32(refr) - xcenter
            offsetc = -_numba.int32(_numba.cuda.threadIdx.x)
            while offsetc < qmax:
                for refc in range(0, Nc):
                    tc = _numba.int32(refc + offsetc)
                    if 0 <= tc < Nc:
                        dqx = qmax + _numba.int32(round(tr2 * tzd[tc] - refr2 * refzd[refc]))
                        dqy = qmax + _numba.int32(round((_numba.float32(tc) - ycenter) * tzd[tc] - (_numba.float32(refc) - ycenter) * refzd[refc]))
                        dqz = maxdqz + _numba.int32(round(cz * (refzd[refc] - tzd[tc])))
                        _numba.cuda.atomic.add(out, (dqz, dqy, dqx), refv[refc] * tv[tc])
                offsetc += _numba.int32(_numba.cuda.blockDim.x)

        def assemble(vals):
            id0, id1, id2 = _numba.cuda.grid(3)
            vals[id0, id1, id2] += vals[-id0, -id1, -id2]
            vals[-id0, -id1, -id2] = vals[id0, id1, id2]

        jkernel = _numba.cuda.jit("void(float32[:,:],float32[:,:],float32[:,:,:])", fastmath=True)(kernel)
        jkernel = jkernel[(1, Nr, qmax), (qmax, 1, 1), stream]
        jassemble = _numba.cuda.jit("void(float32[:,:,:])", fastmath=True)(assemble)
        jassemble = jassemble[(doutput.shape[0], doutput.shape[1], 1), (1, 1, doutput.shape[2] // 2), stream]

        def corr(input):
            """
            Do the correlation
            """
            if stream is None:
                raise ValueError("already closed, use within with statement")
            with stream.auto_synchronize():
                dvals = cuda.to_device(input.astype(_np.float32), stream)
                doutput[:, :, :] = 0
                jkernel(dzd, dvals, doutput)
                jassemble(doutput)
                return doutput.copy_to_host(stream=stream).astype(_np.float64)

        yield corr
    finally:
        if stream is not None:
            stream.synchronize()
        stream = dzd = dvals = doutput = jkernel = jassemble = None
        for x in locals():
            del x
        import gc

        gc.collect()



#!/usr/bin/env python
from os.path import join, exists, dirname, realpath, abspath, isdir
from os import environ, listdir
from os import name as osname
from sys import prefix, path
import setuptools  # noqa # TODO
from distutils.command.sdist import sdist


def configuration():
    from numpy.distutils.misc_util import Configuration
    from numpy.distutils.system_info import (
        get_info,
        default_include_dirs,
        default_lib_dirs,
    )
    from collections import OrderedDict
    import numpy

    config = Configuration("idi", "")
    srcdir = './idi'  # join(dirname(realpath(__file__)), "idi")
    mkl_info = get_info("mkl")
    basedirs = list(
        OrderedDict.fromkeys(
            realpath(p)
            for p in [join(dirname(numpy.__file__), *(4 * [".."])), join(dirname(numpy.__file__), *(3 * [".."])), prefix]
            + [join(*p, *(2 * [".."])) for p in [p.split("site-packages")[:-1] for p in path] if p]
            + [join(*p, "..") for p in [p.split("site-packages")[:-1] for p in path] if p]
            + [join(p, "..") for p in environ["PATH"].split(":")]
        )
    )
    include_dirs = [srcdir]
    library_dirs = default_lib_dirs
    library_dirs.extend(join(b, "lib") for b in basedirs)
    library_dirs.extend(join(b, "lib64") for b in basedirs)
    library_dirs.extend(join(b, "libraries") for b in basedirs)
    library_dirs.extend(join(b, "Library", "lib") for b in basedirs)
    library_dirs.extend(join(b, "Library", "bin") for b in basedirs)

    include_dirs.extend(default_include_dirs)
    include_dirs.extend(join(b, "include") for b in basedirs)
    include_dirs.extend(join(b, "Library", "include") for b in basedirs)

    # print("XXXXXXXXX")
    # print("basedirs", basedirs)
    # print("libdirs:", library_dirs)
    # print("includedirs:", include_dirs)
    # print("np", numpy.__file__)
    # print("env", environ)
    # print("path", path)
    # print("prefix", prefix)
    # print("XXXXXXXXXX")

    include_dirs = [abspath(realpath(p)) for p in filter(isdir, include_dirs)]
    library_dirs = [abspath(realpath(p)) for p in filter(isdir, library_dirs)]

    if mkl_info:
        include_dirs.extend(mkl_info.get("include_dirs"))
        libs = mkl_info.get("libraries", ["mkl_rt"])
    else:
        found_mkl = False
        found_mkl_name = "mkl_rt"
        for d in library_dirs:
            try:
                for f in listdir(d):
                    if f == "mkl_rt.dll" or f == "mkl_rt.so" or f == "mkl_rt.dylib":
                        found_mkl = True
                        found_mkl_name = "mkl_rt"
                    elif "mkl_rt.so." in f and not found_mkl:
                        found_mkl_name = ":" + f
                        found_mkl = True
                if found_mkl:
                    break
            except FileNotFoundError:
                continue
        libs = [found_mkl_name]
        if not osname == "nt":
            libs.extend(["pthread"])

    try:
        from Cython.Build import cythonize

        sources = [join(srcdir, "reconstruction", "autocorrelate3.pyx")]
        have_cython = True
        if not exists(sources[0]):
            print('pyx missing')
            raise FileNotFoundError
    except (ImportError, FileNotFoundError) as e:
        have_cython = False
        sources = [join(srcdir, "reconstruction", "autocorrelate3.c")]
        if not exists(sources[0]):
            print("Cython is required to build ft autocorrelation")
    config.add_extension(
        name="reconstruction.autocorrelate3",
        sources=sources,
        libraries=libs,
        include_dirs=include_dirs,
        library_dirs=library_dirs,
        extra_compile_args=["-DNDEBUG", "-O3", "-DMKL_ILP64"],
    )
    if have_cython:
        config.ext_modules = cythonize(config.ext_modules)
    config.packages.append("idi")
    config.package_dir["idi"] = "./idi"
    config.packages.append("idi.simulation")
    config.package_dir["idi.simulation"] = "./idi/simulation"
    config.packages.append("idi.reconstruction")
    config.package_dir["idi.reconstruction"] = "./idi/reconstruction"
    config.packages.append("idi.util")
    config.package_dir["idi.util"] = "./idi/util"

    return config


def get_version(rel_path):
    import os.path
    import codecs

    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        for line in fp.read().splitlines():
            if line.startswith("__version__"):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


def setup_package():
    try:
        from numpy.distutils.core import setup

        config = configuration().todict()
    except ImportError:
        from setuptools import setup

        config = {"name": "idi"}

    metadata = dict(
        version=get_version("idi/__init__.py"),
        maintainer="zimmf",
        maintainer_email='fzimmermann89+idi@gmail.com',
        description="idi simulation and reconstruction",
        platforms=["Linux", "Mac OS-X"],
        python_requires=">=3.6",
        install_requires=["numpy", "cython", "numba", "numexpr", "scipy", "jinja2", "mkl-service", "matplotlib", "h5py", "mkl", "mkl-include"],
        package_data={"": ["*.cu"]},
        scripts=["scripts/idi_sim.py", "scripts/idi_simrecon.py"],
        test_suite="tests",
        cmdclass={'sdist': sdist},
        **config
    )

    setup(**metadata)

    return None


if __name__ == "__main__":
    setup_package()
